from typing import Dict, Iterator, List

from singer import Transformer, get_logger, metrics, write_record
from singer.utils import strftime, strptime_to_utc

from tap_appsflyer.streams.abstracts import IncrementalStream

LOGGER = get_logger()


class InAppEvents(IncrementalStream):
    tap_stream_id = "in_app_events"
    key_properties = ["event_time", "event_name", "appsflyer_id"]
    replication_keys = ["event_time"]
    path = "api/raw-data/export/app/{}/in_app_events_report/v5"

    def get_url_endpoint(self) -> str:
        return (
            f"{self.client.base_url}/{self.path.format(self.client.config['app_id'])}"
        )



import json
import sys

import singer

from tap_appsflyer.client import Client
from tap_appsflyer.discover import discover
from tap_appsflyer.sync import sync

LOGGER = singer.get_logger()

REQUIRED_CONFIG_KEYS = ["app_id", "api_token"]


def do_discover():
    LOGGER.info("Starting discover")
    catalog = discover()
    json.dump(catalog.to_dict(), sys.stdout, indent=2)
    LOGGER.info("Finished discover")


@singer.utils.handle_top_exception(LOGGER)
def main():
    parsed_args = singer.utils.parse_args(REQUIRED_CONFIG_KEYS)
    state = {}
    if parsed_args.state:
        state = parsed_args.state

    with Client(parsed_args.config) as client:
        if parsed_args.discover:
            do_discover()
        elif parsed_args.catalog:
            sync(
                client=client,
                config=parsed_args.config,
                catalog=parsed_args.catalog,
                state=state,
            )


if __name__ == "__main__":
    main()



from tap_appsflyer.streams.in_app_events import InAppEvents
from tap_appsflyer.streams.installs import Installs
from tap_appsflyer.streams.organic_installs import OrganicInstalls

STREAMS = {
    "installs": Installs,
    "organic_installs": OrganicInstalls,
    "in_app_events": InAppEvents,
}



class appsflyerError(Exception):
    """Class representing Generic Http error."""

    def __init__(self, message=None, response=None):
        super().__init__(message)
        self.message = message
        self.response = response


class appsflyerBackoffError(appsflyerError):
    """Class representing backoff error handling."""

    pass


class appsflyerBadRequestError(appsflyerError):
    """Class representing 400 status code."""

    pass


class appsflyerUnauthorizedError(appsflyerError):
    """Class representing 401 status code."""

    pass


class appsflyerForbiddenError(appsflyerError):
    """Class representing 403 status code."""

    pass


class appsflyerNotFoundError(appsflyerError):
    """Class representing 404 status code."""

    pass


class appsflyerConflictError(appsflyerError):
    """Class representing 406 status code."""

    pass


class appsflyerUnprocessableEntityError(appsflyerBackoffError):
    """Class representing 409 status code."""

    pass


class appsflyerRateLimitError(appsflyerBackoffError):
    """Class representing 429 status code."""

    pass


class appsflyerInternalServerError(appsflyerBackoffError):
    """Class representing 500 status code."""

    pass


class appsflyerNotImplementedError(appsflyerBackoffError):
    """Class representing 501 status code."""

    pass


class appsflyerBadGatewayError(appsflyerBackoffError):
    """Class representing 502 status code."""

    pass


class appsflyerServiceUnavailableError(appsflyerBackoffError):
    """Class representing 503 status code."""

    pass


ERROR_CODE_EXCEPTION_MAPPING = {
    400: {
        "raise_exception": appsflyerBadRequestError,
        "message": "A validation exception has occurred.",
    },
    401: {
        "raise_exception": appsflyerUnauthorizedError,
        "message": "The access token provided is expired, revoked, malformed or invalid for other reasons.",
    },
    403: {
        "raise_exception": appsflyerForbiddenError,
        "message": "You are missing the following required scopes: read",
    },
    404: {
        "raise_exception": appsflyerNotFoundError,
        "message": "The resource you have specified cannot be found.",
    },
    409: {
        "raise_exception": appsflyerConflictError,
        "message": "The API request cannot be completed because the requested operation would conflict with an existing item.",
    },
    422: {
        "raise_exception": appsflyerUnprocessableEntityError,
        "message": "The request content itself is not processable by the server.",
    },
    429: {
        "raise_exception": appsflyerRateLimitError,
        "message": "The API rate limit for your organisation/application pairing has been exceeded.",
    },
    500: {
        "raise_exception": appsflyerInternalServerError,
        "message": "The server encountered an unexpected condition which prevented"
        " it from fulfilling the request.",
    },
    501: {
        "raise_exception": appsflyerNotImplementedError,
        "message": "The server does not support the functionality required to fulfill the request.",
    },
    502: {
        "raise_exception": appsflyerBadGatewayError,
        "message": "Server received an invalid response.",
    },
    503: {
        "raise_exception": appsflyerServiceUnavailableError,
        "message": "API service is currently unavailable.",
    },
}



from typing import Any, Dict, Mapping, Optional, Tuple

import backoff
import requests
from requests import session
from requests.exceptions import ChunkedEncodingError, ConnectionError, Timeout
from singer import get_logger, metrics, utils

from tap_appsflyer.exceptions import (
    ERROR_CODE_EXCEPTION_MAPPING,
    appsflyerBackoffError,
    appsflyerError,
)

LOGGER = get_logger()
REQUEST_TIMEOUT = 300


def raise_for_error(response: requests.Response) -> None:
    """Raises the associated response exception. Takes in a response object,
    checks the status code, and throws the associated exception based on the
    status code.

    :param resp: requests.Response object
    """
    try:
        response_json = response.json()
    except Exception:
        response_json = {}
    if response.status_code != 200:
        if response_json.get("error"):
            message = f"HTTP-error-code: {response.status_code}, Error: {response_json.get('error')}"
        else:
            error_message = response_json.get(
                "message",
                ERROR_CODE_EXCEPTION_MAPPING.get(response.status_code, {}).get(
                    "message", "Unknown Error"
                ),
            )
            message = f"HTTP-error-code: {response.status_code}, Error: {error_message}"

        exc = ERROR_CODE_EXCEPTION_MAPPING.get(response.status_code, {}).get(
            "raise_exception", appsflyerError
        )
        raise exc(message, response) from None


class Client:
    """
    A Wrapper class.
    ~~~
    Performs:
     - Authentication
     - Response parsing
     - HTTP Error handling and retry
    """

    def __init__(self, config: Mapping[str, Any]) -> None:
        self.config = config
        self._session = session()
        self.base_url = "https://hq1.appsflyer.com"

        config_request_timeout = config.get("request_timeout")
        self.request_timeout = (
            float(config_request_timeout) if config_request_timeout else REQUEST_TIMEOUT
        )

    def __enter__(self):
        self.check_api_credentials()
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self._session.close()

    def check_api_credentials(self) -> None:
        pass

    def giveup(self, exc):
        return exc.response is not None and 400 <= exc.response.status_code < 500

    def authenticate(self, headers: Dict, params: Dict) -> Tuple[Dict, Dict]:
        """Authenticates the request with the token."""
        headers["Authorization"] = f"Bearer {self.config.get('api_token')}"
        if "user_agent" in self.config:
            headers["User-Agent"] = self.config["user_agent"]
        return headers, params

    def get(self, endpoint: str, params: Dict, headers: Dict, path: str = None) -> Any:
        """Calls the make_request method with a prefixed method type `GET`"""
        endpoint = endpoint or f"{self.base_url}/{path}"
        headers, params = self.authenticate(headers, params)
        req = requests.Request(
            "GET", endpoint, params=params, headers=headers
        ).prepare()
        return req

    @backoff.on_exception(
        wait_gen=backoff.expo,
        exception=(
            ConnectionResetError,
            ConnectionError,
            ChunkedEncodingError,
            Timeout,
            appsflyerBackoffError,
        ),
        max_tries=5,
        factor=2,
    )
    @backoff.on_exception(
        backoff.expo,
        (requests.exceptions.RequestException),
        max_tries=5,
        giveup=giveup,
        factor=2,
    )
    @utils.ratelimit(10, 1)
    def __make_request(
        self, method: str, endpoint: str, **kwargs
    ) -> Optional[Mapping[Any, Any]]:
        """
        Performs HTTP Operations
        Args:
            method (str): represents the state file for the tap.
            endpoint (str): url of the resource that needs to be fetched
            params (dict): A mapping for url params eg: ?name=Avery&age=3
            headers (dict): A mapping for the headers that need to be sent
            body (dict): only applicable to post request, body of the request

        Returns:
            Dict,List,None: Returns a `Json Parsed` HTTP Response or None if exception
        """
        with metrics.http_request_timer(endpoint) as _:
            response = self._session.request(method, endpoint, **kwargs)
            raise_for_error(response)

        return response.json()



import singer
from singer import metadata
from singer.catalog import Catalog, CatalogEntry, Schema

from tap_appsflyer.schema import get_schemas

LOGGER = singer.get_logger()


def discover() -> Catalog:
    """Run the discovery mode, prepare the catalog file and return the
    catalog."""
    schemas, field_metadata = get_schemas()
    catalog = Catalog([])

    for stream_name, schema_dict in schemas.items():
        try:
            schema = Schema.from_dict(schema_dict)
            mdata = field_metadata[stream_name]
        except Exception as err:
            LOGGER.error(err)
            LOGGER.error(f"stream_name: {stream_name}")
            LOGGER.error(f"type schema_dict: {type(schema_dict)}")
            raise err

        key_properties = metadata.to_map(mdata).get((), {}).get("table-key-properties")

        catalog.streams.append(
            CatalogEntry(
                stream=stream_name,
                tap_stream_id=stream_name,
                key_properties=key_properties,
                schema=schema,
                metadata=mdata,
            )
        )

    return catalog



import csv
import datetime
import re
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Tuple

import pytz
import requests
import singer
from singer import (
    Transformer,
    get_bookmark,
    get_logger,
    metrics,
    utils,
    write_bookmark,
    write_record,
    write_schema,
)
from singer.utils import strftime, strptime_to_utc

LOGGER = get_logger()


SESSION = requests.Session()

# This order matters
fieldnames = (
    "attributed_touch_type",
    "attributed_touch_time",
    "install_time",
    "event_time",
    "event_name",
    "event_value",
    "event_revenue",
    "event_revenue_currency",
    "event_revenue_usd",
    "event_source",
    "is_receipt_validated",
    "af_prt",
    "media_source",
    "af_channel",
    "af_keywords",
    "campaign",
    "af_c_id",
    "af_adset",
    "af_adset_id",
    "af_ad",
    "af_ad_id",
    "af_ad_type",
    "af_siteid",
    "af_sub_siteid",
    "af_sub1",
    "af_sub2",
    "af_sub3",
    "af_sub4",
    "af_sub5",
    "af_cost_model",
    "af_cost_value",
    "af_cost_currency",
    "contributor1_af_prt",
    "contributor1_media_source",
    "contributor1_campaign",
    "contributor1_touch_type",
    "contributor1_touch_time",
    "contributor2_af_prt",
    "contributor2_media_source",
    "contributor2_campaign",
    "contributor2_touch_type",
    "contributor2_touch_time",
    "contributor3_af_prt",
    "contributor3_media_source",
    "contributor3_campaign",
    "contributor3_touch_type",
    "contributor3_touch_time",
    "region",
    "country_code",
    "state",
    "city",
    "postal_code",
    "dma",
    "ip",
    "wifi",
    "operator",
    "carrier",
    "language",
    "appsflyer_id",
    "advertising_id",
    "idfa",
    "android_id",
    "customer_user_id",
    "imei",
    "idfv",
    "platform",
    "device_type",
    "os_version",
    "app_version",
    "sdk_version",
    "app_id",
    "app_name",
    "bundle_id",
    "is_retargeting",
    "retargeting_conversion_type",
    "af_attribution_lookback",
    "af_reengagement_window",
    "is_primary_attribution",
    "user_agent",
    "http_referrer",
    "original_url",
)


class BaseStream(ABC):
    """
    A Base Class providing structure and boilerplate for generic streams
    and required attributes for any kind of stream
    ~~~
    Provides:
     - Basic Attributes (stream_name,replication_method,key_properties)
     - Helper methods for catalog generation
     - `sync` and `get_records` method for performing sync
    """

    url_endpoint = ""
    path = ""
    next_page_key = "next_page"
    headers = {"Accept": "application/json"}

    def __init__(self, client=None) -> None:
        self.client = client
        self.params = {}

    @property
    @abstractmethod
    def tap_stream_id(self) -> str:
        """Unique identifier for the stream.

        This is allowed to be different from the name of the stream, in
        order to allow for sources that have duplicate stream names.
        """

    @property
    @abstractmethod
    def replication_method(self) -> str:
        """Defines the sync mode of a stream."""

    @property
    @abstractmethod
    def replication_keys(self) -> str:
        """Defines the replication key for incremental sync mode of a
        stream."""

    @property
    @abstractmethod
    def forced_replication_method(self) -> str:
        """Defines the sync mode of a stream."""

    @property
    @abstractmethod
    def key_properties(self) -> Tuple[str, str]:
        """List of key properties for stream."""

    @property
    def selected_by_default(self) -> bool:
        """Indicates if a node in the schema should be replicated, if a user
        has not expressed any opinion on whether or not to replicate it."""
        return False

    def get_url_endpoint(self) -> str:
        """Get the URL endpoint for the stream."""
        return self.url_endpoint

    @abstractmethod
    def sync(
        self, state: Dict, schema: Dict, stream_metadata: Dict, transformer: Transformer
    ) -> Dict:
        """Performs a replication sync for the stream.
        ~~~
        Args:
         - state (dict): represents the state file for the tap.
         - schema (dict): Schema of the stream
         - transformer (object): A Object of the singer.transformer class.

        Returns:
         - bool: The return value. True for success, False otherwise.

        Docs:
         - https://github.com/singer-io/getting-started/blob/master/docs/SYNC_MODE.md
        """

    def parse_source_from_url(self, base_url):
        url_regex = re.compile(base_url + r".*/(\w+)_report/v5")
        match = url_regex.match(base_url)
        if match:
            return match.group(1)
        return None

    def get_records(self) -> List:
        """Interacts with api client interaction and pagination."""
        extraction_url = self.url_endpoint

        response = self.client.get(extraction_url, self.params, self.headers)
        if not response:
            LOGGER.warning("No records found in the response")

        with singer.metrics.http_request_timer(
            self.parse_source_from_url(self.client.base_url)
        ) as timer:
            resp = SESSION.send(response)
            timer.tags[singer.metrics.Tag.http_status_code] = resp.status_code
        return resp

    def write_schema(self, schema, stream_name):
        """Write a schema message."""
        try:
            write_schema(stream_name, schema, self.key_properties)
        except OSError as err:
            LOGGER.error(f"OS Error while writing schema for: {stream_name}")
            raise err


class RequestToCsvAdapter:
    def __init__(self, request_data):
        self.request_data_iter = request_data.iter_lines()

    def __iter__(self):
        return self

    def __next__(self):
        return next(self.request_data_iter).decode("utf-8")


class IncrementalStream(BaseStream):
    """Base Class for Incremental Stream."""

    replication_method = "INCREMENTAL"
    forced_replication_method = "INCREMENTAL"
    config_start_key = "start_date"

    @staticmethod
    def get_restricted_start_date(date: str) -> datetime.datetime:
        # https://support.appsflyer.com/hc/en-us/articles/207034366-API-Policy
        restriction_date = utils.now() - datetime.timedelta(days=90)
        start_date = strptime_to_utc(date)

        return max(start_date, restriction_date)

    def get_bookmark(self, state: dict, key: Any = None) -> int:
        """A wrapper for singer.get_bookmark to deal with compatibility for
        bookmark values or start values."""
        get_bookmark_value = get_bookmark(
            state,
            self.tap_stream_id,
            key or self.replication_keys[0],
            self.client.config.get(self.config_start_key, False),
        )

        if get_bookmark_value:
            return self.get_restricted_start_date(get_bookmark_value)

        LOGGER.warning("No bookmark value found, using default start date i.e. 30 days")
        return datetime.datetime.now(pytz.utc) - datetime.timedelta(days=30)

    def get_stop(self, start_datetime, stop_time, days=30):
        if isinstance(start_datetime, datetime.datetime):
            # Ensure we are working with datetime objects
            stop_time = min(start_datetime + datetime.timedelta(days=days), stop_time)
            return stop_time
        else:
            raise TypeError(
                f"Expected start_datetime to be a datetime object, got {type(start_datetime)}"
            )

    def write_bookmark(self, state: dict, key: Any = None, value: Any = None) -> Dict:
        """A wrapper for singer.get_bookmark to deal with compatibility for
        bookmark values or start values."""
        return write_bookmark(
            state, self.tap_stream_id, key or self.replication_keys[0], value
        )

    def xform_boolean_field(self, record, field_name):
        value = record[field_name]
        if value is None:
            return

        record[field_name] = value.strip().lower() == "true"

    def xform_empty_strings_to_none(self, record):
        for key, value in record.items():
            if value == "":
                record[key] = None

    def xform(self, record):
        self.xform_empty_strings_to_none(record)
        self.xform_boolean_field(record, "wifi")
        self.xform_boolean_field(record, "is_retargeting")
        return record

    def sync(
        self, state: Dict, schema: Dict, stream_metadata: Dict, transformer: Transformer
    ) -> Dict:
        self.url_endpoint = self.get_url_endpoint()

        bookmark_date = self.get_bookmark(state)
        from_datetime = bookmark_date

        to_datetime = self.get_stop(from_datetime, datetime.datetime.now(pytz.utc))

        bookmark_date_to_utc = bookmark_date
        current_max_bookmark_date = bookmark_date_to_utc

        self.params["from"] = from_datetime.strftime("%Y-%m-%d %H:%M")
        self.params["to"] = to_datetime.strftime("%Y-%m-%d %H:%M")

        with metrics.record_counter(self.tap_stream_id) as counter:
            request_data = self.get_records()
            csv_data = RequestToCsvAdapter(request_data)
            try:
                reader = csv.DictReader(csv_data, fieldnames)
                next(reader)  # Skip the header row
            except StopIteration:
                LOGGER.warning("No data available in the CSV.")
                return state

            for _, row in enumerate(reader):
                xform_record = self.xform(row)
                transformed_record = transformer.transform(
                    xform_record, schema, stream_metadata
                )
                try:
                    record_timestamp = strptime_to_utc(
                        transformed_record[self.replication_keys[0]]
                    )
                except KeyError as ex:
                    LOGGER.error(
                        "Unable to process Record, Exception occurred: %s for stream %s",
                        ex,
                        self.__class__,
                    )
                    continue
                if record_timestamp >= bookmark_date_to_utc:
                    write_record(self.tap_stream_id, transformed_record)
                    current_max_bookmark_date = max(
                        current_max_bookmark_date, record_timestamp
                    )
                    counter.increment()

            state = self.write_bookmark(
                state, value=strftime(current_max_bookmark_date)
            )
            return counter.value



from typing import Dict

import singer

from tap_appsflyer.client import Client
from tap_appsflyer.streams import STREAMS

LOGGER = singer.get_logger()


def update_currently_syncing(state: Dict, stream_name: str) -> None:
    if not stream_name and singer.get_currently_syncing(state):
        del state["currently_syncing"]
    else:
        singer.set_currently_syncing(state, stream_name)
    singer.write_state(state)


def sync(client: Client, config: Dict, catalog: singer.Catalog, state) -> None:
    """Sync selected streams from catalog."""

    selected_streams = []
    for stream in catalog.get_selected_streams(state):
        selected_streams.append(stream.stream)
    LOGGER.info(f"selected_streams: {selected_streams}")

    last_stream = singer.get_currently_syncing(state)
    LOGGER.info(f"last/currently syncing stream: {last_stream}")

    with singer.Transformer() as transformer:
        for stream_name in selected_streams:

            stream = STREAMS[stream_name](client)
            stream_catalog = catalog.get_stream(stream_name)
            stream_schema = stream_catalog.schema.to_dict()
            stream_metadata = singer.metadata.to_map(stream_catalog.metadata)

            stream.write_schema(stream_schema, stream_name)

            LOGGER.info(f"START Syncing: {stream_name}")
            update_currently_syncing(state, stream_name)
            total_records = stream.sync(
                state=state,
                schema=stream_schema,
                stream_metadata=stream_metadata,
                transformer=transformer,
            )

            update_currently_syncing(state, None)
            LOGGER.info(
                f"FINISHED Syncing: {stream_name}, total_records: {total_records}"
            )



from setuptools import setup, find_packages

setup(
    name="tap-appsflyer",
    version="1.0.0",
    description="Singer.io tap for extracting data from appsflyer API",
    author="Stitch",
    url="http://singer.io",
    classifiers=["Programming Language :: Python :: 3 :: Only"],
    py_modules=["tap_appsflyer"],
    install_requires=[
        "attrs==25.1.0",
        "singer-python==6.1.0",
        "requests==2.32.5",
        "backoff==2.2.1",
    ],
    entry_points="""
        [console_scripts]
        tap-appsflyer=tap_appsflyer:main
    """,
    packages=find_packages(),
    package_data={
        "tap_appsflyer": ["schemas/*.json"],
    },
    include_package_data=True,
)



import json
import os
from typing import Dict, Tuple

import singer
from singer import metadata

from tap_appsflyer.streams import STREAMS

LOGGER = singer.get_logger()


def get_abs_path(path: str) -> str:
    """Get the absolute path for the schema files."""
    return os.path.join(os.path.dirname(os.path.realpath(__file__)), path)


def load_schema_references() -> Dict:
    """Load the schema files from the schema folder and return the schema
    references."""
    shared_schema_path = get_abs_path("schemas/shared")

    shared_file_names = []
    if os.path.exists(shared_schema_path):
        shared_file_names = [
            f
            for f in os.listdir(shared_schema_path)
            if os.path.isfile(os.path.join(shared_schema_path, f))
        ]

    refs = {}
    for shared_schema_file in shared_file_names:
        with open(os.path.join(shared_schema_path, shared_schema_file)) as data_file:
            refs["shared/" + shared_schema_file] = json.load(data_file)

    return refs


def get_schemas() -> Tuple[Dict, Dict]:
    """Load the schema references, prepare metadata for each streams and return
    schema and metadata for the catalog."""
    schemas = {}
    field_metadata = {}

    refs = load_schema_references()
    for stream_name, stream_obj in STREAMS.items():
        schema_path = get_abs_path(f"schemas/{stream_name}.json")
        with open(schema_path) as file:
            schema = json.load(file)

        schemas[stream_name] = schema
        schema = singer.resolve_schema_references(schema, refs)

        mdata = metadata.new()
        mdata = metadata.get_standard_metadata(
            schema=schema,
            key_properties=getattr(stream_obj, "key_properties"),
            valid_replication_keys=(getattr(stream_obj, "replication_keys") or []),
            replication_method=getattr(stream_obj, "replication_method"),
        )
        mdata = metadata.to_map(mdata)

        automatic_keys = getattr(stream_obj, "replication_keys") or []
        for field_name in schema["properties"].keys():
            if field_name in automatic_keys:
                mdata = metadata.write(
                    mdata, ("properties", field_name), "inclusion", "automatic"
                )

        mdata = metadata.to_list(mdata)
        field_metadata[stream_name] = mdata

    return schemas, field_metadata
